//###################################################################################
// File: Counter.C
// Schneller genauer Frequenzzhler fr Frequenzen <100Hz
// Theoretisch bis zu 1.8mHz bei einer Quarzfrequenz von 8MHz
//
// Das zu messende Signal wird am INT0 Pin eingespeist.
// 
// Messart:   Periodendauermessung mit Kehrwertbildung
// Messzeit:  maximal 2 * Periodendauer
//            0.1Hz -> 20s
//            1Hz   -> 2s
//            10Hz  -> 0.2s  (+Anzeigezeit von ca. 200ms)
//            100Hz -> 0.02s (+Anzeigezeit von ca. 200ms)
//            usw.
//
//            Frequenzen >100Hz lassen sich auch messen, aber die
//            Auflsung ist dann nicht besonders gut.
//
//
// Weitere Anwendungsmglichkeiten evtl. Drehzahlmessung
// 
//###################################################################################
// Compiler: AVR-GCC (Mingw Win32)
// Version : ????
// sprintf(); aus gcctest9.c (etwas gendert)
//###################################################################################
// Processor: AT90S8515
//
// hk@holger-klabunde.de
// http://www.holger-klabunde.de
// Last change: 08.01.2002
//###################################################################################

#include <io.h>
#include <interrupt.h>
#include <sig-avr.h>
#include <stdarg.h>
#include <string.h>

#include "protos.h"

#define NOP() asm volatile ("nop" ::)

//Fr sprintf(); um da nicht alles zu ndern
typedef unsigned char  u08;
typedef          char  s08;
typedef unsigned short u16;
typedef          short s16;
typedef u08            bool;

int sprintf(u08 *buf, const u08 *format, ...);

//"volatile" fr alle Variablen die in Interrupt's gendert werden
unsigned char volatile ready;
unsigned int volatile overflow;
unsigned int volatile over;
unsigned int volatile count,count1,count2; //Lesen Timer1 bei erster und zweiter Flanke
unsigned long ticks,frequ;
char line[43];            //Puffer fr sprintf(line,...);
double period, frequency;

#define TAKT 8     //Taktfrequenz in MHz
#define PRESCALE 1 //Prescaler Timer1

#define RED_ON() { cbi(PORTD,7); }
#define RED_OFF() { sbi(PORTD,7); }

#define GREEN_ON() { cbi(PORTD,6); }
#define GREEN_OFF() { sbi(PORTD,6); }

const double fscale=1.0E6*TAKT/PRESCALE;

//###################################################################################
INTERRUPT(SIG_INTERRUPT0)        /* signal handler for INT0 interrupt */
//###################################################################################
{
//neu
 count=inw(TCNT1L); //Timer1 lesen
//alt
// count=__inw(TCNT1L); //Timer1 lesen

 if(ready==0) //Erste Flanke gefunden
  {
   count1=count; //Startwert merken
   overflow=0;   //TMR1 berlufe lschen bei erster Flanke
  }

 if(ready==1) //Zweite Flanke gefunden
  {
   count2=count; //Endwert merken
   over=overflow;     //TMR1 berlufe retten !
  }

 ready++; //Flankenzhler
}

//###################################################################################
INTERRUPT(SIG_OVERFLOW1)        /* signal handler for tcnt1 overflow interrupt */
//###################################################################################
{
 overflow++; //Timer1 berlufe zhlen
}

//###################################################################################
/* Hauptprogramm */
int main(void)
//###################################################################################
{
 unsigned char i;
 
 outp(0xFF,DDRC);  // set io-pins on PortC
 outp(0x00,PORTC);  //

 outp(0xFF,DDRA);  // set io-pins on PortA
 outp(0x00,PORTA);  //

 outp(0xF0,DDRD);  // set io-pins on PortD
 outp(0x00,PORTD);  //

 Delay1ms(20); 
 LCDInit();
 ser_init();

//Int0
 outp((1<<ISC01)|(1<<ISC00), MCUCR); //Int bei Flanke 0->1
 outp((1<<INT0), GIMSK); /* enable external int0 */

//Timer1
 outp(0x00,TCCR1A); //
 outp(0x01,TCCR1B); //Timer1 Start Takt = F/1
 outp((1<<TOIE1), TIMSK); /* enable timer1 overflow int */

 LCDCls();
 LCDWrite("Counter");

 sei();             //enable interrupts
 ser_puts("Counter\n");

 overflow=0;
 i=0;

 for(;;) // Anfang Programm Endlosschleife
  {

   cli();             // disable interrupts
   ready=0;           //Neue Messung starten
   outp(0xff,GIFR);    //alle aufgelaufenen Interrupts lschen
   outp(0xff,TIFR);    //alle aufgelaufenen Interrupts lschen
   NOP(); NOP();
//   outw(0x0000,TCNT1L); //geht nicht mehr
//   __outw(0,TCNT1L);

   outp(0x00,TCNT1L); //Dann eben so
   outp(0x00,TCNT1H); //Dann eben so
   outp(0x01,TCCR1B); //Timer1 Start Takt = F/1
   sei();             //enable interrupts
   NOP(); NOP();
   RED_ON(); GREEN_OFF();

//Das hier geht nur wenn ready als "volatile" deklariert wird !
//ready wird im Interrupt hochgezhlt. Ohne volatile wird das hier
//vom Optimizer wegoptimiert und funktioniert einfach nicht !
   while(ready<2) //Warten auf Messung
    {
     NOP();
    }

   i++;
   outp(i,PORTA);

   cli();             // disable interrupts
   outp(0x00,TCCR1B); //Timer1 Stop
   sei();             //enable interrupts

   //Jetzt gibt es zwei Mglichkeiten
   //1. Timer1 Overflows=0  Ticks = Endwert - Startwert
   //2. Timer1 Overflows>=1 Ticks = (65536 - Startwert) + ((Overflow-1) * 65536) + Endwert
   if(over==0) ticks=(unsigned long)(count2-count1);
   else
    {
     ticks=(unsigned long)(65536-count1);
     over--;
     ticks+=over*(unsigned long)65536;
     ticks+=(unsigned long)count2;
    }

   period=(double)ticks/fscale;      //Periodendauer
   frequency=(double)1000.0/period;   //Frequenz in 0.001Hz
   frequ=(unsigned long)frequency;
   ticks=ticks/TAKT;                 //Periode: Ticks in us umrechnen


//Periodendauer anzeigen
   RED_OFF(); GREEN_ON();
   LCDPos(1,1); //10
   if(ticks>=1000000) //Anzeige in Sekunden
    {
     line[0]=0; //Sonst rger mit dem strcat() in sprintf()
     sprintf(line,"%u.",ticks/1000000);
     LCDWrite(line);
     ser_puts(line);
     ticks=ticks%1000000;
     if(ticks<100000) { LCDWriteByte('0'); ser_putc('0'); }
     if(ticks<10000) { LCDWriteByte('0'); ser_putc('0'); }
     if(ticks<1000) { LCDWriteByte('0'); ser_putc('0'); }
     if(ticks<100) { LCDWriteByte('0'); ser_putc('0'); }
     if(ticks<10) { LCDWriteByte('0'); ser_putc('0'); }
     line[0]=0; //Sonst rger mit dem strcat() in sprintf()
     sprintf(line,"%us ",ticks);
     LCDWrite(line);
     ser_puts(line);
    }
   else if(ticks>=1000) //Anzeige in ms
    {
     LCDWriteByte(' ');  ser_putc(' ');
     if(ticks<100000) { LCDWriteByte(' '); ser_putc(' '); }
     if(ticks<10000) { LCDWriteByte(' '); ser_putc(' '); }
     line[0]=0; //Sonst rger mit dem strcat() in sprintf()
     sprintf(line,"%u.",ticks/1000);
     LCDWrite(line);
     ser_puts(line);
     ticks=ticks%1000;
     if(ticks<100) { LCDWriteByte('0'); ser_putc('0'); }
     if(ticks<10) { LCDWriteByte('0'); ser_putc('0'); }
     line[0]=0; //Sonst rger mit dem strcat() in sprintf()
     sprintf(line,"%ums",ticks);
     LCDWrite(line);
     ser_puts(line);
    }
   else //Anzeige in us
    {
     line[0]=0; //Sonst rger mit dem strcat() in sprintf()
     sprintf(line,"     %uus",ticks);
     LCDWrite(line);
     ser_puts(line);
    }

//Frequenz anzeigen
//vier Stellen vor und drei Stellen nach dem Komma
   LCDPos(2,1); //10
   if(frequ<1000000) { LCDWriteByte(' ');  ser_putc(' '); } //Fhrende Spaces
   if(frequ<100000) { LCDWriteByte(' '); ser_putc(' '); }
   if(frequ<10000) { LCDWriteByte(' '); ser_putc(' '); }
   line[0]=0; //Sonst rger mit dem strcat() in sprintf()
   sprintf(line,"%u.",(frequ/1000));
   LCDWrite(line);
   ser_puts(line);
   frequ=frequ%1000;
   if(frequ<100) { LCDWriteByte('0'); ser_putc('0'); }
   if(frequ<10) { LCDWriteByte('0'); ser_putc('0'); }
   line[0]=0; //Sonst rger mit dem strcat() in sprintf()
   sprintf(line,"%uHz ",frequ);
   LCDWrite(line);
   ser_puts(line);
   ser_puts("\n");
 
   Delay1ms(200);  //Anzeige ein bichen stehen lassen !
 
  } // Ende Programm Endlosschleife
}

#define SCRATCH 16


//###################################################################################
int sprintf(u08 *buf, const u08 *format, ...)
//###################################################################################
// simplified sprintf
{
  u08 scratch[SCRATCH];
  u08 format_flag;
  u16 base; //vorher auch u_val=0
  u08 *ptr;
  va_list ap;
  unsigned long u_val=0; //gibt warnings

  va_start (ap, format);
  for (;;){
    while ((format_flag = *format++) != '%'){      // Until '%' or '\0'
      if (!format_flag){va_end (ap); return (0);}
      *buf = format_flag; buf++; *buf=0;
    }

    switch (format_flag = *format++){

    default:
      *buf = format_flag; buf++; *buf=0;
      continue;
    case 'i':
      if (((int)u_val) < 0){
        u_val = - u_val;
        *buf = '-'; buf++; *buf=0;
      }
      // no break -> run into next case
    case 'u':
      base = 10;
      goto CONVERSION_LOOP;
    case 'x':
      base = 16;

    CONVERSION_LOOP:
//      u_val = va_arg(ap,int);
      u_val = va_arg(ap,long); //gibt warnings
      ptr = scratch + SCRATCH;
      *--ptr = 0;
      do
       {
        char ch = u_val % base + '0';
        if (ch > '9') ch += 'a' - '9' - 1;
        *--ptr = ch;
        u_val /= base;
       } while (u_val);

      strcat(buf, ptr);
      buf += strlen(ptr);
    }
  }
}


//###################################################################################
//Nicht sonderlich genau !!
void Delay1ms(unsigned int time)
//###################################################################################
{
 unsigned int i;

 for(i=0; i<time; i++) Delay100us(10);
}

//###################################################################################
//Nicht sonderlich genau !!
//time<=255 !!
void Delay100us(unsigned char time)
//###################################################################################
{
 unsigned char k,l;

 for(k=0; k<time; k++)
  {
   for(l=0; l<20; l++)
    {
     //Ein CPU-Clock = 0.1us bei 10MHz
     NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP();  
     NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP();  
     NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP();  
     NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP();  
     NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP();  
    }
  }
}
